using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Test
{
    public class Hello
    {
        public void HelloWorld()
        {
            Console.WriteLine("Hello World!");
        }
    }
}
