# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import pluggy

hookimpl = pluggy.HookimplMarker("surfactant")
